package eu.jonahbauer.raytracing.scene.transform;

import eu.jonahbauer.raytracing.math.AABB;
import eu.jonahbauer.raytracing.math.Ray;
import eu.jonahbauer.raytracing.math.Vec3;
import eu.jonahbauer.raytracing.scene.HitResult;
import eu.jonahbauer.raytracing.scene.Hittable;
import eu.jonahbauer.raytracing.scene.Target;
import org.jetbrains.annotations.NotNull;

import java.util.random.RandomGenerator;

public sealed class RotateY extends Transform {
    private final double cos;
    private final double sin;

    private final @NotNull AABB bbox;

    public static @NotNull RotateY create(@NotNull Hittable object, double angle) {
        if (object instanceof Target) {
            return new RotateYTarget(object, angle);
        } else {
            return new RotateY(object, angle);
        }
    }

    private RotateY(@NotNull Hittable object, double angle) {
        super(object);
        this.cos = Math.cos(angle);
        this.sin = Math.sin(angle);

        var bbox = object.getBoundingBox();

        var min = new Vec3(Double.MAX_VALUE, Double.MAX_VALUE, Double.MAX_VALUE);
        var max = new Vec3(- Double.MAX_VALUE, - Double.MAX_VALUE, - Double.MAX_VALUE);

        for (int i = 0; i < 2; i++) {
            for (int j = 0; j < 2; j++) {
                for (int k = 0; k < 2; k++) {
                    var x = i * bbox.max().x() + (1 - i) * bbox.min().x();
                    var y = j * bbox.max().y() + (1 - j) * bbox.min().y();
                    var z = k * bbox.max().z() + (1 - k) * bbox.min().z();

                    var newx =  cos * x + sin * z;
                    var newz = -sin * x + cos * z;

                    var temp = new Vec3(newx, y, newz);

                    min = Vec3.min(min, temp);
                    max = Vec3.max(max, temp);
                }
            }
        }

        this.bbox = new AABB(min, max);
    }

    @Override
    protected final @NotNull Ray transform(@NotNull Ray ray) {
        var origin = ray.origin();
        var direction = ray.direction();

        var newOrigin = transform(origin);
        var newDirection = transform(direction);
        return new Ray(newOrigin, newDirection);
    }

    @Override
    protected final @NotNull HitResult transform(@NotNull HitResult result) {
        var position = result.position();
        var newPosition = untransform(position);

        var normal = result.normal();
        var newNormal = untransform(normal);

        return result.withPositionAndNormal(newPosition, newNormal);
    }

    protected final @NotNull Vec3 transform(@NotNull Vec3 vec) {
        return new Vec3(cos * vec.x() - sin * vec.z(), vec.y(), sin * vec.x() + cos * vec.z());
    }

    protected final @NotNull Vec3 untransform(@NotNull Vec3 vec) {
        return new Vec3(cos * vec.x() + sin * vec.z(), vec.y(), - sin * vec.x() + cos * vec.z());
    }

    @Override
    public @NotNull AABB getBoundingBox() {
        return bbox;
    }

    @Override
    public @NotNull String toString() {
        return object + " rotated by " + Math.toDegrees(Math.atan2(sin, cos)) + "° around the y axis";
    }

    private static final class RotateYTarget extends RotateY implements Target {

        private RotateYTarget(@NotNull Hittable object, double angle) {
            super(object, angle);
        }

        @Override
        public double getProbabilityDensity(@NotNull Vec3 origin, @NotNull Vec3 direction) {
            return ((Target) object).getProbabilityDensity(transform(origin), transform(direction));
        }

        @Override
        public @NotNull Vec3 getTargetingDirection(@NotNull Vec3 origin, @NotNull RandomGenerator random) {
            return untransform(((Target) object).getTargetingDirection(transform(origin), random));
        }
    }
}
